# -*- coding: utf-8 -*-
#
# Copyright 2014 Jaime Gil de Sagredo Luna
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The `models` module contains the `booby` highest level abstraction:
the `Model`.

To define a model you should subclass the :class:`Model` class and
add a list of :mod:`fields` as attributes. And then you could instantiate
your `Model` and work with these objects.

Something like this::

    class Repo(Model):
         name = fields.String()
         owner = fields.Embedded(User)

    booby = Repo(
        name='Booby',
        owner={
            'login': 'jaimegildesagredo',
            'name': 'Jaime Gil de Sagredo'
        })

    print booby.to_json()
    '{"owner": {"login": "jaimegildesagredo", "name": "Jaime Gil de Sagredo"}, "name": "Booby"}'
"""

import six
import collections
try:
    import ujson as json
except ImportError:
    import json


from booby import mixins, fields, errors, _utils


class ModelMeta(type):
    def __new__(cls, name, bases, attrs):
        attrs['_fields'] = {}

        for base in bases:
            if hasattr(base, '_fields'):
                for k, v in base._fields.items():
                    attrs['_fields'][k] = v
            for k, v in base.__dict__.items():
                if isinstance(v, fields.Field):
                    attrs['_fields'][k] = v

        for k, v in attrs.items():
            if isinstance(v, fields.Field):
                attrs['_fields'][k] = v

        return super(ModelMeta, cls).__new__(cls, name, bases, attrs)

    def __repr__(cls):
        return '<{}.{}({})>'.format(cls.__module__, cls.__name__,
                                    _utils.repr_options(cls._fields))


@six.add_metaclass(ModelMeta)
class Model(mixins.Encoder):
    """The `Model` class. All Booby models should subclass this.

    By default the `Model's` :func:`__init__` takes a list of keyword arguments
    to initialize the `fields` values. If any of these keys is not a `field`
    then raises :class:`errors.FieldError`. Of course you can overwrite the
    `Model's` :func:`__init__` to get a custom behavior.

    You can get or set Model `fields` values in two different ways: through
    object attributes or dict-like items::

        >>> booby.name is booby['name']
        True
        >>> booby['name'] = 'booby'
        >>> booby['foo'] = 'bar'
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        errors.FieldError: foo

    :param \*\*kwargs: Keyword arguments with the fields values to initialize the model.

    """

    def __new__(cls, *args, **kwargs):
        model = super(Model, cls).__new__(cls)
        model._data = {}

        return model

    def __init__(self, **kwargs):
        self._update(kwargs)

    def __repr__(self):
        cls = type(self)

        return '<{}.{}({})>'.format(cls.__module__, cls.__name__,
                                    _utils.repr_options(dict(self)))

    def __iter__(self):
        for name in self._fields:
            value = getattr(self, name)

            if isinstance(value, Model):
                value = dict(value)
            elif isinstance(value, collections.MutableSequence):
                value = self._encode_sequence(value)

            yield name, value

    def _encode_sequence(self, sequence):
        result = []

        for value in sequence:
            if isinstance(value, Model):
                value = dict(value)

            result.append(value)

        return result

    def __getitem__(self, k):
        if k not in self._fields:
            raise errors.FieldError(k)

        return getattr(self, k)

    def __setitem__(self, k, v):
        if k not in self._fields:
            raise errors.FieldError(k)

        setattr(self, k, v)

    def update(self, *args, **kwargs):
        """This method updates the `model` fields values with the given `dict`.
        The model can be updated passing a dict object or keyword arguments,
        like the Python's builtin :py:func:`dict.update`.

        """

        self._update(dict(*args, **kwargs))

    def _update(self, values):
        for k, v in values.items():
            self[k] = v

    @property
    def is_valid(self):
        """This property will be `True` if there are not validation
        errors in this `model` fields. If there are any error then
        will be `False`.

        This property wraps the :func:`Model.validate` method to be
        used in a boolean context.

        """

        try:
            self.validate()
        except errors.ValidationError:
            return False
        else:
            return True

    @property
    def descriptions(self):
        """This property returns a dict of strings with the name of propery as a key
        and their description as a value"""

        res = {}
        for name, field in list(self._fields.items()):
            res[name] = field.description
        return res

    def description(self, key):
        try:
            return self._fields[key].description
        except KeyError:
            return None
            
    def validate(self):
        """This method validates the entire `model`. That is, validates
        all the :mod:`fields` within this model.

        If some `field` validation fails, then this method raises the same
        exception that the :func:`field.validate` method had raised, but
        with the field name prepended.

        """

        for name, field in self._fields.items():
            try:
                field.validate(getattr(self, name))
            except errors.ValidationError as err:
                raise errors.ValidationError('%s %s' % (name, err))

    @property
    def validation_errors(self):
        """Generator of field name and validation error string pairs
        for each validation error on this `model` fields.

        """

        for name, field in self._fields.items():
            try:
                field.validate(getattr(self, name))
            except errors.ValidationError as err:
                yield name, str(err)

    def to_json(self, *args, **kwargs):
        """This method returns the `model` as a `json string`. It receives
        the same arguments as the builtin :py:func:`json.dump` function.

        To build a json representation of this `model` this method iterates
        over the object to build a `dict` and then serializes it as json.

        """

        return json.dumps(dict(self), *args, **kwargs)

    @classmethod
    def decode(self, raw):
        result = {}

        for name, field in self._fields.items():
            try:
                value = raw[field.options.get('name', name)]
            except KeyError:
                continue
            else:
                value = field.decode(value)

            result[name] = value

        return result

    @classmethod
    def properties(cls):
        """Get properties defined in Model without instantiate them"""
        if hasattr(cls, "_fields"):
            return cls._fields
        else:
            ret = {}
            for k, v in six.iteritems(cls.__dict__):
                if k.startswith("_"):
                    continue
            
                ret[k] = v
    
        return ret
    
__all__ = ("Model", )




# -*- coding: utf-8 -*-
#
# Copyright 2014 Jaime Gil de Sagredo Luna
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from . import mixins, errors
from .helpers import nullable


class Encoder(object):
    def __call__(self, value):
        return self.encode(value)


class Model(Encoder):
    @nullable
    def encode(self, value):
        return value.encode()


class List(Encoder):
    def __init__(self, *encoders):
        self._encoders = encoders

    @nullable
    def encode(self, value):
        if not isinstance(value, collections.MutableSequence):
            raise errors.EncodeError()

        result = []
        for item in value:
            for encoder in self._encoders:
                item = encoder(item)

            result.append(item)

        return result


class DateTime(Encoder):
    def __init__(self, format=None):
        self._format = format

    @nullable
    def encode(self, value):
        if self._format is None:
            return value.isoformat()

        return value.strftime(self._format)


class Collection(List):
    def __init__(self):
        super(Collection, self).__init__(Model())



# -*- coding: utf-8 -*-
#
# Copyright 2014 Jaime Gil de Sagredo Luna
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The `validators` module contains a set of :mod:`fields` validators.

A validator is any callable `object` which receives a `value` as the
target for the validation. If the validation fails then should raise an
:class:`errors.ValidationError` exception with an error message.

`Validators` are passed to :class:`fields.Field` and subclasses as possitional
arguments.

"""

import re

import six
import collections
import datetime

from booby import errors
from booby.helpers import nullable


class Validator(object):
    def __call__(self, value):
        self.validate(value)

    def validate(self, value):
        raise NotImplementedError()


class Required(Validator):
    """This validator forces fields to have a value other than :keyword:`None`."""

    def validate(self, value):
        if value is None:
            raise errors.ValidationError('is required')


class In(Validator):
    """This validator forces fields to have their value in the given list.

    :param choices: A `list` of possible values.

    """

    def __init__(self, choices):
        self.choices = choices

    def validate(self, value):
        if value not in self.choices:
            raise errors.ValidationError('should be in {}'.format(self.choices))


class String(Validator):
    """This validator forces fields values to be an instance of `basestring`."""

    @nullable
    def validate(self, value):
        if not isinstance(value, six.string_types):
            raise errors.ValidationError('should be a string')


class Integer(Validator):
    """This validator forces fields values to be an instance of `int`."""

    @nullable
    def validate(self, value):
        if not isinstance(value, six.integer_types):
            raise errors.ValidationError('should be an integer')


class Float(Validator):
    """This validator forces fields values to be an instance of `float`."""

    @nullable
    def validate(self, value):
        if not isinstance(value, float):
            raise errors.ValidationError('should be a float')


class Boolean(Validator):
    """This validator forces fields values to be an instance of `bool`."""

    @nullable
    def validate(self, value):
        if not isinstance(value, bool):
            raise errors.ValidationError('should be a boolean')


class Model(Validator):
    """This validator forces fields values to be an instance of the given
    :class:`models.Model` subclass and also performs a validation in the
    entire `model` object.

    :param model: A subclass of :class:`models.Model`

    """

    def __init__(self, model):
        self.model = model

    @nullable
    def validate(self, value):
        if not isinstance(value, self.model):
            raise errors.ValidationError(
                "should be an instance of '{}'".format(self.model.__name__))

        value.validate()


class Email(String):
    """This validator forces fields values to be strings and match a
    valid email address.

    """

    def __init__(self):
        super(Email, self).__init__()

        self.pattern = re.compile('^[^@]+\@[^@]+$')

    @nullable
    def validate(self, value):
        super(Email, self).validate(value)

        if self.pattern.match(value) is None:
            raise errors.ValidationError('should be a valid email')


class URL(String):
    """This validator forces fields values to be strings and match a
    valid URL.

    """

    def __init__(self):
        super(URL, self).__init__()

        # From Django validator:
        #   https://github.com/django/django/blob/master/django/core/validators.py#L47
        self.pattern = re.compile(
            r'^https?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    @nullable
    def validate(self, value):
        super(URL, self).validate(value)

        if self.pattern.match(value) is None:
            raise errors.ValidationError('should be a valid URL')


class IP(String):
    """This validator forces fields values to be strings and match a
    valid IP address.

    """

    def __init__(self):
        super(IP, self).__init__()

        # Regex from:
        #   https://gist.github.com/mnordhoff/2213179
        self.ipv4 = re.compile(r'^(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')
        self.ipv6 = re.compile(r'^(?:(?:[0-9A-Fa-f]{1,4}:){6}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|::(?:[0-9A-Fa-f]{1,4}:){5}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){4}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){3}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,2}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:){2}(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,3}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}:(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,4}[0-9A-Fa-f]{1,4})?::(?:[0-9A-Fa-f]{1,4}:[0-9A-Fa-f]{1,4}|(?:(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.){3}(?:[0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5]))|(?:(?:[0-9A-Fa-f]{1,4}:){,5}[0-9A-Fa-f]{1,4})?::[0-9A-Fa-f]{1,4}|(?:(?:[0-9A-Fa-f]{1,4}:){,6}[0-9A-Fa-f]{1,4})?::)$')

    @nullable
    def validate(self, value):
        super(IP, self).validate(value)

        if self.ipv4.match(value) is None and self.ipv6 is None:
            raise errors.ValidationError('should be a valid IPv4 or IPv6 address')


class URI(String):
    """This validator forces fields values to be strings and match a
    valid URI.

    """

    def __init__(self):
        super(URI, self).__init__()

        self.uri_regex = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')

    @nullable
    def validate(self, value):
        super(URI, self).validate(value)

        if self.uri_regex.match(value) is None:
            raise errors.ValidationError('should be a valid URI')
        

class List(Validator):
    """This validator forces field values to be a :keyword:`list`.
    Also a list of inner :mod:`validators` could be specified to validate
    each list element. For example, to validate a list of
    :class:`models.Model` you could do::

        books = fields.Field(validators.List(validators.Model(YourBookModel)))

    :param \*validators: A list of inner validators as possitional arguments.

    """

    def __init__(self, *validators):
        self.validators = validators

    @nullable
    def validate(self, value):
        if not isinstance(value, collections.MutableSequence):
            raise errors.ValidationError('should be a list')

        for i in value:
            for validator in self.validators:
                validator(i)


class DateTime(Validator):
    @nullable
    def validate(self, value):
        if not isinstance(value, datetime.datetime):
            raise errors.ValidationError('should be a datetime')



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

# Import requirements
with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='booby-ng',
    version='0.8.2-2',
    install_requires=required,
    description='Data modeling and validation Python library',
    long_description=long_description,
    url='https://github.com/cr0hn/booby',
    author='Jaime Gil de Sagredo Luna',
    author_email='jaimegildesagredo@gmail.com',
    maintainer='Daniel Garcia (cr0hn)',
    maintainer_email='cr0hn@cr0hn.com',
    packages=find_packages(exclude=['tests', 'tests.*']),
    extras_require={
        'ustraJSON': ["ujson"],
    },
    use_2to3=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )



# -*- coding: utf-8 -*-
#
# Copyright 2014 Jaime Gil de Sagredo Luna
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The `errors` module contains all exceptions used by Booby."""


class BoobyError(Exception):
    """Base class for all Booby exceptions."""

    pass


class FieldError(BoobyError):
    """This exception is used as an equivalent to :class:`AttributeError`
    for :mod:`fields`.

    """

    pass


class ValidationError(BoobyError):
    """This exception should be raised when a `value` doesn't validate.
    See :mod:`validators`.

    """

    pass


class EncodeError(BoobyError):
    pass


class DecodeError(BoobyError):
    pass



# -*- coding: utf-8 -*-
#
# Copyright 2014 Jaime Gil de Sagredo Luna
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The :mod:`inspection` module provides users and 3rd-party library
developers a public api to access :mod:`booby` objects and classes internal
data, such as defined fields, and some low-level type validations.

This module is based on the Python :py:mod:`inspect` module.

"""

from booby import models


def get_fields(model):
    """Returns a `dict` mapping the given `model` field names to their
    `fields.Field` objects.

    :param model: The `models.Model` subclass or instance you want to
                  get their fields.

    :raises: :py:exc:`TypeError` if the given `model` is not a model.

    """

    if not is_model(model):
        raise TypeError(
            '{} is not a {} subclass or instance'.format(model, models.Model))

    return dict(model._fields)


def is_model(obj):
    """Returns `True` if the given object is a `models.Model` instance
    or subclass. If not then returns `False`.

    """

    try:
        return (isinstance(obj, models.Model) or
                issubclass(obj, models.Model))

    except TypeError:
        return False
